(define-record-type Association
  (fields key value))

(env-add! (base-module) '=> (fenius-rtd-class (record-type-descriptor Association)))

(define class-Dict (fenius-rtd-class (record-type-descriptor hashmap)))

(define (make-dict . assocs)
  (let loop ([hmap empty-hashmap] [assocs assocs])
    (if (null? assocs)
        hmap
      (loop (hashmap-add hmap (Association-key (car assocs))
                              (Association-value (car assocs)))
            (cdr assocs)))))

(define (fenius-write-hashmap hmap port)
  (display "dict(" port)
  (let ([first? #t])
    (hashmap-for-each (lambda (key value)
                        (cond [first? (set! first? #f)]
                              [else (display ", " port)])
                        (fenius-write key port)
                        (display " => " port)
                        (fenius-write value port))
                      hmap))
  (display ")" port))

(define *not-found* (gensym "not-found"))

(add-method! class-Dict 'at
  (lambda (dict)
    (lambda (key)
      (let ([result (hashmap-ref dict key *not-found*)])
        (if (eq? result *not-found*)
            (error 'dict/at "Element not found")
          result)))))

(add-method! class-Dict 'get
  (lambda (dict)
    (case-lambda
      [(key) (dict-get dict key *not-found*)]
      [(key default) (dict-get dict key default)])))

(define (dict-get dict key default)
  (let ([result (hashmap-ref dict key default)])
    (if (eq? result *not-found*)
        (error 'dict/at "Element not found")
      result)))

(add-method! class-Dict 'has_key
  (lambda (dict)
    (lambda (key)
      (not (eq? (hashmap-ref dict key *not-found*) *not-found*)))))

(add-method! class-Dict 'with
  (lambda (dict)
    (lambda items
      (fold-left (lambda (dict item)
                   (hashmap-add dict (Association-key item) (Association-value item)))
                 dict
                 items))))

(add-method! class-Dict 'without
  (lambda (dict)
    (lambda items
      (fold-left (lambda (dict key)
                   (hashmap-remove dict key))
                 dict
                 items))))


(env-add! (base-module) 'dict make-dict)
