(define & '&)

(define-syntax match
  (lambda (form)
    (lambda (get-property)
      (syntax-case form ()
        [(_ expr clauses ...)

         (let ()

           (define (gen-match-clauses val clauses)
             (syntax-case clauses ()
               [() #'(error 'match "No clause matching ~s" val)]
               [([pattern body ...] others ...)
                #`(let ([fail (lambda () #,(gen-match-clauses val #'(others ...)))])
                    #,(gen-match-clause #'val #'pattern #'(begin body ...) #'fail))]
               [other (syntax-violation 'match "Malformed match clause: ~s" #'other)]))

           (define (gen-match-clause val pattern body fail)
             (syntax-case pattern (list quote &)
               [id #;where (identifier? #'id)
                           #`(let ([id #,val])
                               #,body )]

               [(quote datum)
                #`(if (equal? #,val 'datum)
                      #,body
                    (#,fail))]

               [(list)
                #`(if (null? #,val)
                      #,body
                    (#,fail))]

               [(list & rest)
                (gen-match-clause val #'rest body fail)]

               [(list pat rest ...)
                #`(if (pair? #,val)
                      #,(gen-match-clause
                         #`(car #,val)
                         #'pat
                         (gen-match-clause #`(cdr #,val) #'(list rest ...) body fail)
                         fail)
                    (#,fail))]

               [(head pats ...)
                (let ([prop (get-property #'head)])
                  (if (record-property? prop)
                      #`(if ((record-predicate (record-type-descriptor head)) #,val)
                            #,(gen-match-record
                               (rtd-accessors (record-property-rtd prop) #'head)
                               val
                               #'(pats ...)
                               body
                               fail)
                          (#,fail))
                    (syntax-violation 'match "Don't know how to match head ~s" #'head)))]

               [other (syntax-violation 'match "Don't know how to match pattern ~s" #'other)]))
           (define (record-property? prop)
             (and (pair? prop)
                  (pair? (cdr prop))
                  (record-type-descriptor? (cadr prop))))

           (define (record-property-rtd prop)
             (cadr prop))

           (define (gen-match-record accessors val pats body fail)
             (syntax-case pats ()
               [() (if (null? accessors)
                       body
                     (syntax-violation 'match "Fewer patterns than record fields" pats))]
               [(pat others ...)
                (if (null? accessors)
                    (syntax-violation 'match "More patterns than record fields" pats)
                  (gen-match-clause
                   #`(#,(car accessors) #,val)
                   #'pat
                   (gen-match-record (cdr accessors) val #'(others ...) body fail)
                   fail))]))

           (define (rtd-accessors rtd context)
             (if (not rtd)
                 '()
               (let* ([fields (record-type-field-names rtd)]
                      [accessors (let loop ([i 0])
                                   (if (>= i (vector-length fields))
                                       '()
                                     (cons #`(record-accessor (record-type-descriptor
                                                               #,(datum->syntax context (record-type-name rtd)))
                                                              #,i)
                                           (loop (+ i 1)))))])
                 (append (rtd-accessors (record-type-parent rtd) context)
                         accessors))))

           ;;(define (gen-match-record rtd val pats body fail)



           #`(let ([val expr])
               #,(gen-match-clauses #'val #'(clauses ...))))]))))
