(define fenius-version "0.3")

(print-gensym #f)

;;;;
;;;; Main program
;;;;

(env-add! (base-module) 'true #t)
(env-add! (base-module) 'false #f)
(env-add! (base-module) 'nil nil)
(env-add! (base-module) 'let specialop-let)
(env-add! (base-module) 'method specialop-method)
(env-add! (base-module) 'fn specialop-fn)
(env-add! (base-module) 'if specialop-if)
(env-add! (base-module) 'while specialop-while)
(env-add! (base-module) 'do specialop-do)
(env-add! (base-module) 'redo specialop-redo)
(env-add! (base-module) 'return specialop-return)
(env-add! (base-module) 'quote specialop-quote)
(env-add! (base-module) 'record specialop-record)
(env-add! (base-module) 'import specialop-import)
(env-add! (base-module) 'try specialop-try)
(env-add! (base-module) 'raise specialop-raise)
(env-add! (base-module) ':= specialop-assignment)
(env-add! (base-module) (string->symbol ".") specialop-dot)

(env-add! (base-module) (string->symbol "..")
          (lambda (start end)
            (make-Range start end #f)))

(env-add! (base-module) (string->symbol "...")
          (lambda (start end)
            (make-Range start end #t)))

(env-add! (base-module) (string->symbol "..+")
          (lambda (start length)
            (make-Range start (+ start length) #f)))

(env-add! (base-module) (string->symbol "...+")
          (lambda (start length)
            (make-Range start (+ start length) #t)))


(env-add! (base-module) 'print (lambda args
                                 (for-each fenius-display args)
                                 (newline)))
(env-add! (base-module) (string->symbol "++")
          (lambda (left right)
            (cond [(string? left) (string-append left right)]
                  [(list? left) (append left right)]
                  [else (error (string->symbol "++")
                               "Cannot append objects" left right)])))
(env-add! (base-module) '+ +)
(env-add! (base-module) '- -)
(env-add! (base-module) '* *)
(env-add! (base-module) '/ /)
(env-add! (base-module) '== equal?)
(env-add! (base-module) '!= (lambda (x y) (not (equal? x y))))
(env-add! (base-module) '< <)
(env-add! (base-module) '> >)
(env-add! (base-module) '<= <=)
(env-add! (base-module) '>= >=)
(env-add! (base-module) 'exit exit)

(env-add! (base-module) 'Object class-Object)
(env-add! (base-module) 'List class-List)
(env-add! (base-module) 'Number class-Number)
(env-add! (base-module) 'Tuple class-Tuple)
(env-add! (base-module) 'String class-String)

;; Let's have some meta fun here.
(env-add! (base-module) 'Phrase (fenius-rtd-class (record-type-descriptor Phrase)))
(env-add! (base-module) 'Constant (fenius-rtd-class (record-type-descriptor Constant)))
(env-add! (base-module) 'Identifier (fenius-rtd-class (record-type-descriptor Identifier)))
(env-add! (base-module) 'Tuple (fenius-rtd-class (record-type-descriptor Tuple)))
(env-add! (base-module) 'Array (fenius-rtd-class (record-type-descriptor Array)))
(env-add! (base-module) 'Body (fenius-rtd-class (record-type-descriptor Body)))
(env-add! (base-module) 'Block (fenius-rtd-class (record-type-descriptor Block)))
(env-add! (base-module) 'Call (fenius-rtd-class (record-type-descriptor Call)))
(env-add! (base-module) 'Index (fenius-rtd-class (record-type-descriptor Index)))
(env-add! (base-module) 'Macro (fenius-rtd-class (record-type-descriptor Macro)))

(env-add! (base-module) 'evaluate evaluate)
(env-add! (base-module) 'baseModule (base-module))

(define user-module (build-module 'std/user '()))

(define (main progname . args)
  (cond [(null? args) (display "Use `-i` to run the REPL, or call with a filename to run.\n")]
        [(equal? (car args) "-i") (fenius-repl)]
        [else (run-fenius-program (car args) (cdr args))]))

(define (run-fenius-program filename args)
  (load-fenius-module-from-file "main" filename))

(define (repl-read-program)
  (let loop ([lines '()])
    (display (if (null? lines) "fenius> " "     "))
    (flush-output-port (current-output-port))
    (let* ([line (get-line (current-input-port))])
      (if (eof-object? line)
          (if (null? lines)
              (begin (display "Bye!\n") (exit))
            (error 'repl-read-program "^D"))
        (let* ([lines (cons line lines)]
               [text (string-join (reverse lines) "\n")]
               [source (build-source "<stdin>" text)])
          (guard (e [(source-eof-error? e) (loop lines)])
            (let*-values ([(body rest) (parse-body (make-location source 0)
                                                   (parse-source-tokens source))])
              (if (null? rest)
                  body
                (error 'repl-read-program "Extra tokens: " rest)))))))))


(define (fenius-repl)
  (printf "Welcome to Fenius ~a (~a)! Type 'exit()' or ^D to quit.\n"
          fenius-version (current-git-revision))
  (let loop ()
    (guard (e [#t (handle-repl-error e)])
      (let* ([body (repl-read-program)]
             [result (run-fenius-body body user-module)])
        (fenius-write result (current-output-port))
        (newline)))
    (loop)))

(define (handle-repl-error err)
  (display "Interpreter exception: ")
  (display (if (message-condition? err)
               (condition-message err)
             err))
  (when (irritants-condition? err)
    (display " ")
    (write (condition-irritants err)))
  (newline))

;; Apparently, in R6RS all calls have to be after all definitions...

(register-module! (base-module))
(env-add! (base-module) 'base (base-module))
(module-import-contents! user-module (base-module))


(apply main (command-line))
