(define (fenius-match pattern value env)
  (match pattern
    [(Identifier _ '_) '()]
    [(Identifier _ name) (list (make-binding name value))]
    [(Constant _ const) (if (equal? const value)
                            '()
                          #f)]
    [(Call _ head* args)
     (let ([head (evaluate head* env)])
       (cond [(and (Class? head) (Class-rtd head))
              (let ([fields (record-type-all-accessors (Class-rtd head))])
                (cond [(fenius-instance-of? value head)
                       (cond [(= (length args) (length fields))
                              (let ([field-vals (map (lambda (field-descriptor)
                                                       ((cadr field-descriptor) value))
                                                     fields)])
                                (fenius-match-all args field-vals env))]
                             [else (error 'fenius-match "Number of pattern arguments does not match constructor for class" head)])]
                      [else #f]))]
             [(Macro? head)
              (fenius-match ((Macro-transformer head) pattern) value env)]
             [else (error 'fenius-match "Don't know how to match head" head)]))]
    [(Array _ patterns)
     (cond [(list? value)
            (fenius-match-all patterns value env)]
           [else #f])]
    [else (error 'fenius-match "Don't know how to match pattern" pattern)]))


(define (fenius-match-all patterns values env)
  (let loop ([bindings '()] [patterns patterns] [values values])
    (match patterns
      ['() (cond [(null? values) bindings]
                 [else #f])]
      [(list (Call _ (Identifier _ '*) (list pat)) & rest)
       (when (not (null? rest))
         (error 'fenius-match-all "Star pattern must be the last one in the list"))
       (cond [(fenius-match pat values env)
              => (lambda (result)
                   (append result bindings))]
             [else #f])]
      [(list pat & rest)
       (cond [(fenius-match pat (car values) env)
              => (lambda (result)
                   (loop (append result bindings) rest (cdr values)))]
             [else #f])]
      )))



;; Can 'obviously' refactor to use reflection.
(define (meta-expand ast transform keep-loc?)
  (define wild (make-Identifier #f '_))
  (define reloc (if keep-loc?
                    (lambda (loc) (make-Constant #f loc))
                  (lambda (loc) wild)))
  (define (sub items)
    (map (lambda (ast) (meta-expand ast transform keep-loc?)) items))
  (define (meta-list items)
    (make-Array #f items))
  (or (transform ast)
      (match ast
        [(Identifier loc name)
         (make-Call #f (make-Identifier #f 'Identifier)
                    (list (reloc loc) (make-Constant #f name)))]
        [(Constant loc value)
         (make-Call #f (make-Identifier #f 'Constant)
                    (list (reloc loc) (make-Constant #f value)))]
        [(Call loc head args)
         (make-Call #f (make-Identifier #f 'Call)
                    (list (reloc loc)
                          (meta-expand head transform keep-loc?)
                          (meta-list (sub args))))]
        [(Index loc head args)
         (make-Call #f (make-Identifier #f 'Index)
                    (list (reloc loc)
                          (meta-expand head transform keep-loc?)
                          (meta-list (sub args))))]
        [(Block loc body)
         (make-Call #f (make-Identifier #f 'Block)
                    (list (reloc loc) (meta-expand body transform keep-loc?)))]
        [(Body loc phrases)
         (make-Call #f (make-Identifier #f 'Body)
                    (list (reloc loc) (meta-list (sub phrases))))]
        [(Phrase loc constituents)
         (make-Call #f (make-Identifier #f 'Phrase)
                    (list (reloc loc) (meta-list (sub constituents))))]
        [(Array loc items)
         (make-Call #f (make-Identifier #f 'Array)
                    (list (reloc loc) (meta-list (sub items))))]
        [(Tuple loc items)
         (make-Call #f (make-Identifier #f 'Tuple)
                    (list (reloc loc) (meta-list (sub items))))])))

(define (meta-escape ast)
  (match ast
    [(Call _ (Identifier _ '_) (list expr)) expr]
    [else #f]))

;; (env-add! (base-module) 'meta
;;   (make-Macro
;;    (lambda (form)
;;      (match form
;;        [(Call _ _meta (list keep-loc-arg expr))
;;         (let ([keep-loc? (match keep-loc-arg
;;                            [(Identifier _ 'true) #t]
;;                            [(Identifier _ 'false) #f]
;;                            [_ (error 'meta "keep-loc? must be literal true or false")])])
;;           (meta-expand expr meta-escape keep-loc?))]))))


(env-add! (base-module) 'ast_match
  (make-Macro
   (lambda (form)
     (match form
       [(Call _ _match (list expr))
        (meta-expand expr meta-escape #f)]))))

(env-add! (base-module) 'ast_gen
  (make-Macro
   (lambda (form)
     (match form
       [(Call _ _gen (list expr))
        (meta-expand expr meta-escape #t)]))))


(define specialop-match
  (make-SpecialOperator
   (lambda (form env)
     (match form
       [(Phrase _ (list _match expr (Block _ (Body _ phrases))))
        (let ([clauses (parse-match-clauses phrases)]
              [value (evaluate expr env)])
          (let loop ([clauses clauses])
            (match clauses
              ['() (error 'specialop-match "No matching clause in match form")]
              [(list (list pattern result) & rest)
               (cond [(fenius-match pattern value env)
                      => (lambda (bindings)
                           (evaluate result (build-env bindings env)))]
                     [else (loop rest)])])))]))))

(define (parse-match-clauses clauses)
  (match clauses
    ['() '()]
    [(list (Call _ (Identifier _ '=>) (list pattern result)) & rest)
     (cons (list pattern result)
           (parse-match-clauses rest))]))


(env-add! (base-module) 'match specialop-match)









(define (replace-ast-location ast loc)
  (define (sub items)
    (map (lambda (ast) (replace-ast-location ast loc)) items))
  (match ast
    [(Identifier _ name) (make-Identifier loc name)]
    [(Constant _ value) (make-Constant loc value)]
    [(Call _ head args) (make-Call loc head (sub args))]
    [(Index _ head args) (make-Index loc head (sub args))]
    [(Phrase _ constituents) (make-Phrase loc (sub constituents))]
    [(Block _ body) (make-Block loc (replace-ast-location body loc))]
    [(Array _ items) (make-Array loc (sub items))]
    [(Tuple _ items) (make-Tuple loc (sub items))]
    [(Body _ phrases) (make-Body loc (sub phrases))]))
