(define (record->list rec)
  ;; For debugging. :(
  (cond [(record? rec)
         (let* ([rtd (record-rtd rec)]
                [name (record-type-name rtd)]
                [numfields (vector-length (record-type-field-names rtd))])
           (list->vector
            (cons name (let loop ([i 0])
                         (if (>= i numfields)
                             '()
                           (cons (record->list ((record-accessor rtd i) rec))
                                 (loop (+ i 1))))))))]
        [(list? rec) (map record->list rec)]
        [else rec]))

(define (string-index char text start)
  ;; Return the index of the first occurence of `char` in `text`, beginning at `start`.
  ;; If the character is not present, return the position after the last character.
  (let ([len (string-length text)])
    (let loop ([i start])
      (cond [(>= i len) len] ;; makes sense for parse-token
            [(eqv? char (string-ref text i)) i]
            [else (loop (+ i 1))]))))

(define (string-join items separator)
  (if (null? items)
      ""
    (let loop ([text (car items)] [items (cdr items)])
      (cond [(null? items) text]
            [else (loop (string-append text separator (car items))
                        (cdr items))]))))

(define (string-split string separator)
  (let ([len (string-length string)])
    (let loop ([parts '()] [start 0])
      (cond [(>= start len) (reverse parts)]
            [else (let ([end (string-index separator string start)])
                    (loop (cons (substring string start end) parts)
                          (+ 1 end)))]))))

(define-syntax current-git-revision
  ;; XXX: Is there a better way to compute a value at compile-time?
  (lambda (form)
    (syntax-case form ()
      [(_) (let* ([ports (process "git rev-parse --short HEAD 2>/dev/null")]
                  [revision (get-line (car ports))])
             (close-port (car ports))
             (close-port (cadr ports))
             (datum->syntax #'here (if (eof-object? revision)
                                       #f
                                     revision)))])))
